# Box, file versioning.
# Copyright (C) 2023  Firlast
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from os import path as os_path, walk as os_walk


def _filter_filelist(path: str) -> str:
    if path.startswith('/'):
        path = path[1:]

    return os_path.abspath(path)


def _path_has_ignored(ignored: list, path: str) -> bool:
    for i in ignored:
        if i in os_path.abspath(path):
            return True


def _load_ignore() -> list:
    try:
        with open('.ignore', 'r') as ignore_file:
            filelist = ignore_file.readlines()
    except FileNotFoundError:
        ignored = []
    else:
        filelist = [file.replace('\n', '') for file in filelist]
        ignored = list(map(_filter_filelist, filelist))

    ignored.append('.box')
    return ignored


def get_non_ignored() -> list:
    ignored = _load_ignore()
    non_ignored = []

    for root, dirs, files in os_walk('.'):
        if not _path_has_ignored(ignored, root):
            for file in files:
                filepath = os_path.join(root, file)
                if not _path_has_ignored(ignored, filepath):
                    non_ignored.append(filepath.replace('./', ''))

    return non_ignored



import os
import sys
import json
import shutil
import marshal

import bupytest

sys.path.insert(0, './')

from box import tracker
from box import commit

REPO_DIR = '.box'
OBJECT_DIR = os.path.join(REPO_DIR, 'objects')
FILE_TESTS_DIR = os.path.join('tests', 'files')

if os.path.isdir(FILE_TESTS_DIR):
    shutil.rmtree(FILE_TESTS_DIR, ignore_errors=True)

if os.path.isdir(REPO_DIR):
    shutil.rmtree(REPO_DIR, ignore_errors=True)

os.mkdir(FILE_TESTS_DIR)
os.mkdir(REPO_DIR)
os.mkdir(OBJECT_DIR)

TEST_FILE_1 = os.path.join(FILE_TESTS_DIR, 'readme.md')
TEST_FILE_2 = os.path.join(FILE_TESTS_DIR, 'hello.txt')

TEST_FILE_1_CONTENT = (
    '# Box Version Control Test\n'
    'This is a test!'
)

TEST_FILE_1_CONTENT_CHANGED = (
    '# Box Version Control Test\n'
    'This is a test!\n'
    'A big test!'
)

TEST_FILE_2_CONTENT = (
    'Hello Word!\n'
    'How are you?'
)

TEST_FILE_2_CONTENT_CHANGED = (
    'Hello Word!\n'
)

with open(TEST_FILE_1, 'w') as file:
    file.write(TEST_FILE_1_CONTENT)

with open(TEST_FILE_2, 'w') as file:
    file.write(TEST_FILE_2_CONTENT)


_tracker = tracker.Tracker()
_commit = commit.Commit()


class TestTracker(bupytest.UnitTest):
    def __init__(self):
        super().__init__()

    def test_track(self):
        _tracker.track([TEST_FILE_1])
        tracked = _tracker.get_tracked()

        self.assert_true(len(tracked) == 1, message='Tracked data larger than necessary')
        self.assert_true(tracked.get(TEST_FILE_1), message='File not found in tracked data')
        self.assert_expected(
            value=tracked[TEST_FILE_1]['committed'],
            expected=False,
            message='Committed status is not False'
        )

        self._tracked_file_hash = tracked[TEST_FILE_1]['hash']
        self._tracked_file_committed = tracked[TEST_FILE_1]['committed']

    def test_track_other_file(self):
        _tracker.track([TEST_FILE_2])
        tracked = _tracker.get_tracked()

        self.assert_true(len(tracked) == 2, message='Tracked data larger (or smaller) than necessary')
        self.assert_true(tracked.get(TEST_FILE_2), message='File not found in tracked data')
        self.assert_expected(
            value=tracked[TEST_FILE_2]['committed'],
            expected=False,
            message='Committed status is not False'
        )

    def test_update_committed_status(self):
        _tracker.update_track_info(TEST_FILE_1, committed=True)
        tracked = _tracker.get_tracked()

        self.assert_expected(
            value=tracked[TEST_FILE_1]['hash'],
            expected=self._tracked_file_hash,
            message='Hash updated unnecessarily'
        )

        self.assert_expected(
            value=tracked[TEST_FILE_1]['committed'],
            expected=True,
            message='Committed status not updated'
        )


class TestCommit(bupytest.UnitTest):
    def __init__(self):
        super().__init__()

    def test_get_commits(self):
        commits = _commit.get_commits()
        self.assert_expected(commits, {})

    def test_commit(self):
        _tracker.update_track_info(TEST_FILE_1, committed=False)
        commit_id = _commit.commit('author', 'email', [TEST_FILE_1, TEST_FILE_2], message='first commit')
        commits = _commit.get_commits()

        self.assert_expected(len(commits), 1, message='Incorrect commits number')
        self.assert_true(commits.get(commit_id), message='Commit ID not found')
        self.assert_true(commits[commit_id]['message'], 'first commit')
        self.assert_expected(len(commits[commit_id]['objects']), 2, message='Expected 2 objects references')

        file_1_object_path = os.path.join(OBJECT_DIR, commits[commit_id]['objects'][TEST_FILE_1])
        file_2_object_path = os.path.join(OBJECT_DIR, commits[commit_id]['objects'][TEST_FILE_2])

        self.assert_true(
            value=os.path.isfile(file_1_object_path),
            message=f'Object to {repr(TEST_FILE_1)} not created'
        )

        self.assert_true(
            value=os.path.isfile(file_2_object_path),
            message=f'Object to {repr(TEST_FILE_2)} not created'
        )

    def test_commit_changed_file(self):
        with open(TEST_FILE_1, 'w') as file_1:
            file_1.write(TEST_FILE_1_CONTENT_CHANGED)

        with open(TEST_FILE_2, 'w') as file_2:
            file_2.write(TEST_FILE_2_CONTENT_CHANGED)

        commit_id = _commit.commit('author', 'email', [TEST_FILE_1, TEST_FILE_2], message='second commit')
        commits = _commit.get_commits()

        self.assert_expected(len(commits), 2, message='Incorrect commits number')
        self.assert_true(commits.get(commit_id), message='Commit ID not found')
        self.assert_true(commits[commit_id]['message'], 'second commit')
        self.assert_expected(len(commits[commit_id]['objects']), 2, message='Expected 1 object reference')

        file_1_object_path = os.path.join(OBJECT_DIR, commits[commit_id]['objects'][TEST_FILE_1])
        file_2_object_path = os.path.join(OBJECT_DIR, commits[commit_id]['objects'][TEST_FILE_2])

        self.assert_true(
            value=os.path.isfile(file_1_object_path),
            message=f'Object to {repr(TEST_FILE_1)} not created'
        )

        self.assert_true(
            value=os.path.isfile(file_2_object_path),
            message=f'Object to {repr(TEST_FILE_2)} not created'
        )

        with open(file_1_object_path, 'rb') as object_file:
            file_1_object_lines = marshal.load(object_file)

        with open(file_2_object_path, 'rb') as object_file:
            file_2_object_lines = marshal.load(object_file)

        self.assert_expected(
            value=file_1_object_lines,
            expected={
                1: 'This is a test!\n',
                2: 'A big test!'
            }
        )

        # "None" is a deleted line
        self.assert_expected(
            value=file_2_object_lines,
            expected={
                1: None
            }
        )


class TestIntegrity(bupytest.UnitTest):
    def __init__(self):
        super().__init__()

    def test_check_integrity(self):
        self.assert_true(_commit.check_integrity())

    def test_external_change_object(self):
        commits = _commit.get_commits()
        last_commit = list(commits.values())[-1]
        commit_objects = last_commit['objects']

        file_1_object_id = commit_objects[TEST_FILE_1]
        file_1_object_path = os.path.join(OBJECT_DIR, file_1_object_id)

        with open(file_1_object_path, 'rb') as obj:
            original_content = obj.read()

        with open(file_1_object_path, 'wb') as obj:
            obj.write(b'changed content')

        self.assert_false(_commit.check_integrity())

        # rewrite original content to object
        with open(file_1_object_path, 'wb') as obj:
            obj.write(original_content)

    def test_commit_data_change_detect(self):
        commits = _commit.get_commits()
        last_commit = list(commits.keys())[-1]

        commits[last_commit]['author'] = 'Anonymous User'
        
        with open(os.path.join(REPO_DIR, 'commits.json'), 'w') as file:
            json.dump(commits, file)

        self.assert_false(_commit.check_integrity())



# Box, file versioning.
# Copyright (C) 2023  Firlast
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import json
import marshal
from hashlib import md5, sha1
from os import path
from typing import List
from datetime import datetime

from .tracker import Tracker
from . import exceptions
from . import utils


class Commit:
    def __init__(self) -> None:
        """
        This class handles everything about `commits`.
        """

        repo_path = '.box'

        self._commit_file = path.join(repo_path, 'commits.json')
        self._obj_file = path.join(repo_path, 'objects')
        self._tracker = Tracker()

        self._commits = self._get_commits()

    def _dump_commit_file(self, commits: dict) -> None:
        self._commits = commits
        with open(self._commit_file, 'w') as file:
            json.dump(commits, file, indent=2)

    def _create_object(self, file_diff: dict, obj_id: str) -> None:
        object_path = path.join(self._obj_file, obj_id)

        with open(object_path, 'wb') as _object:
            marshal.dump(file_diff, _object)

    def _create_object_to_binary(self, file_content: bytes, obj_id: str) -> None:
        object_path = path.join(self._obj_file, obj_id)

        with open(object_path, 'wb') as _object:
            _object.write(file_content)

    def _get_object(self, object_id: str) -> dict:
        object_path = path.join(self._obj_file, object_id)
        with open(object_path, 'rb') as file:
            object_data = marshal.load(file)

        return object_data

    def _get_object_hash(self, object_id: str) -> bytes:
        object_path = path.join(self._obj_file, object_id)
        with open(object_path, 'rb') as file:
            _hash = md5(file.read())

        return _hash.digest()

    def _get_file_commits(self, filename: str) -> dict:
        commits = self.get_commits()
        file_commits = {cid: cdata for cid, cdata in commits.items() if filename in cdata['objects']}
        return file_commits

    def _get_commits(self) -> dict:
        try:
            with open(self._commit_file) as file:
                commits = json.load(file)
        except FileNotFoundError:
            commits = {}

        return commits

    def _get_commit_hash(self, commit: dict) -> str:
        commit = commit.copy()
        objects_hashes = [self._get_object_hash(obj) for obj in commit.pop('objects').values()]
        objects_sum_hash = md5(b''.join(objects_hashes)).hexdigest()
        commit_info = '.'.join([objects_sum_hash, *commit.values()])
        _hash = md5(commit_info.encode()).hexdigest()

        return _hash

    def _get_last_commit_hash(self) -> str:
        commits = self.get_commits()

        if commits:
            _hash = self._get_commit_hash(list(commits.values())[-1])
        else:
            _hash = ''

        return _hash

    def _merge_lines(self, objects: list) -> dict:
        merged = {}
        for obj in objects:
            merged.update(obj)

        return merged

    def get_commits(self, until_commit_id: str = None) -> dict:
        """
        Get all commits in `dict` format. The commit
        data contains author name and email, commit datetime,
        message and objects references.

        :param until_commit_id: Get all commits until a commit ID.
        :return: Commits
        """

        if until_commit_id:
            filtered_commits = {}
            for cid, cdata in self._commits.items():
                if cid != until_commit_id:
                    filtered_commits[cid] = cdata
                else:
                    filtered_commits[cid] = cdata
                    break

            return filtered_commits

        return self._commits

    def merge_objects(self, file: str) -> dict:
        """Merge all objects from file. The result is a
        dictionary with enumerated lines.

        :param file: File to merge
        :type file: str
        :return: Merged objects
        :rtype: dict
        """

        file_commits = self._get_file_commits(file)
        file_objects = [self._get_object(commit['objects'][file]) for commit in file_commits.values()]
        return self._merge_lines(file_objects)

    def _create_commit_objects(self, files: list, _id: str) -> dict:
        tracked = self._tracker.get_tracked()
        commit_objects = {}

        for file in files:
            if file not in tracked:
                raise exceptions.FileNotTrackedError(f'File "{file}" not tracked')

        for file in files:
            obj_id = utils.generate_id(_id, file)
            commit_objects[file] = obj_id
            file_info = tracked[file]

            if file_info['binary']:
                current_hash = self._tracker.get_file_hash(file)
                tracked[file]['hash'] = current_hash
                with open(file, 'rb') as file_r:
                    file_content = file_r.read()

                if file_info['committed']:
                    if tracked[file]['hash'] != current_hash:
                        self._create_object_to_binary(file_content, obj_id)
                else:
                    tracked[file]['committed'] = True
                    self._create_object_to_binary(file_content, obj_id)
            else:
                with open(file, 'r') as file_r:
                    file_lines = utils.enumerate_lines(file_r.readlines())

                if not file_info['committed']:
                    tracked[file]['committed'] = True
                else:
                    current_hash = self._tracker.get_file_hash(file)

                    if file_info['hash'] != current_hash:
                        merged = self.merge_objects(file)
                        tracked[file]['hash'] = current_hash
                        file_lines = utils.difference_lines(merged, file_lines)

                self._create_object(file_lines, obj_id)

        self._tracker.dump_tracker(tracked)
        return commit_objects

    def commit(self, author: str, author_email: str, files: List[str], message: str) -> str:
        """Commit files with a message.

        If is the first file commit, this method enumerate
        all file lines and store this in an "object", the
        object ID is stored in commit data. Otherwise, this method
        gets the difference of all merged file commits from
        enumerate file lines and store this difference.

        :param author: Author name
        :type author: str
        :param author_email: Author email
        :type author_email: str
        :param files: Files to commit
        :type files: List[str]
        :param message: Message to describe changes
        :type message: str
        :raises exceptions.NoFilesToCommitError: If no file has changed
        :return: Return commit ID
        :rtype: str
        """

        commits = self.get_commits()

        commit_datetime = str(datetime.now().replace(microsecond=0))
        _id_parts = ''.join((author, author_email, message, commit_datetime))
        commit_objects = self._create_commit_objects(files, _id_parts)

        if not commit_objects:
            raise exceptions.NoFilesToCommitError('No files to commit')

        commit_data = dict(
            author=author,
            author_email=author_email,
            message=message,
            date=str(commit_datetime),
            objects=commit_objects
        )

        _last_hash = self._get_last_commit_hash()
        _commit_id_parts = '.'.join((self._get_commit_hash(commit_data), _last_hash))
        commit_id = sha1(_commit_id_parts.encode()).hexdigest()
        commits[commit_id] = commit_data

        self._dump_commit_file(commits)
        return commit_id

    def check_integrity(self) -> bool:
        """Check commits integrity.

        If any commit is altered by a third party, `False` is
        returned, indicating that the commit chain is invalid.

        :return: If chain is valid
        :rtype: bool
        """

        commits = self.get_commits()
        last_hash = ''

        for commit_id in commits.keys():
            c_hash = self._get_commit_hash(commits[commit_id])
            _hash_parts = '.'.join((c_hash, last_hash)).encode()
            sum_hash = sha1(_hash_parts).hexdigest()
            if commit_id == sum_hash:
                last_hash = c_hash
            else:
                return False
        
        return True



# Box, file versioning.
# Copyright (C) 2023  Firlast
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from hashlib import sha1
from secrets import token_hex


def enumerate_lines(file_lines: list) -> dict:
    lines = enumerate(file_lines)
    return dict((ln, l) for ln, l in lines)


def difference_lines(older_enum_lines: dict, new_enum_lines: dict) -> dict:
    difference = {}

    # get deleted or existent lines difference
    for number, line in older_enum_lines.items():
        new_line = new_enum_lines.get(number)

        if not new_line:
            difference[number] = None
        elif new_line != line:
            difference[number] = new_line

    # get new lines
    for number, line in new_enum_lines.items():
        if not older_enum_lines.get(number):
            difference[number] = line

    return difference


def generate_id(*complements: str) -> str:
    id_parts = ''.join((
        token_hex(16),
        *complements
    ))
    return sha1(id_parts.encode()).hexdigest()


def divide_list(parts: int, iterable: list) -> list:
    if parts > 1:
        q, r = divmod(len(iterable), parts)
        split_list = []
        stop = 0

        for i in range(1, parts + 1):
            start = stop
            stop += q + 1 if i <= r else q
            split_list.append(iterable[start:stop])

        return split_list
    else:
        raise ValueError('"parts" must be at least 1')



# Box, file versioning.
# Copyright (C) 2023  Firlast
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys
import time
from os import path, makedirs
from typing import Union

from argeasy import ArgEasy

from .__init__ import __version__
from .tracker import Tracker
from .commit import Commit
from .ignore import get_non_ignored
from . import _config
from . import exceptions
from . import utils

REPO_PATH = '.box'
OBJECTS_PATH = path.join(REPO_PATH, 'objects')

tracker = Tracker()
commit = Commit()


def _get_uncommitted_files(tracked: dict) -> list:
    uncommitted = [filepath for filepath, info in tracked.items() if not info['committed']]
    changed_files = [filepath for filepath, info in tracked.items() if info['hash'] != tracker.get_file_hash(filepath)]
    return (*uncommitted, *changed_files)


def _get_untracked_files(non_ignored: list, tracked: dict) -> list:
    return [file for file in non_ignored if file not in tracked]


def _init() -> None:
    if path.isdir(REPO_PATH):
        print(f'\033[33mRepository already started in {repr(REPO_PATH)}\033[m')
        sys.exit(1)
    else:
        makedirs(OBJECTS_PATH)
        print(f'\033[32mNew repository started in {repr(REPO_PATH)}\033[m')


def _add(files: Union[list, str]) -> None:
    if files == "*":
        non_ignored = get_non_ignored()
        tracked = tracker.get_tracked()
        files = _get_untracked_files(non_ignored, tracked)

    for file in files:
        if not path.isfile(file):
            print(f'\033[31mFile {repr(file)} not exists in current directory\033[m')
            sys.exit(1)

    tracker.track(files)


def _status() -> None:
    non_ignored = get_non_ignored()
    tracked_files = tracker.get_tracked()

    uncommitted = _get_uncommitted_files(tracked_files)
    untracked = _get_untracked_files(non_ignored, tracked_files)

    if uncommitted or untracked:
        print('\033[1mUncommitted files\033[m')

        if uncommitted:
            print(f'Use "commit" argument to commit {len(uncommitted)} files:\033[33m')
            for file in uncommitted:
                print('   ' + file)
        else:
            print('0 files found for commit')

        print('\n\033[37;1mUntracked files\033[m')

        if untracked:
            print(f'Use "add" argument to track {len(untracked)} files:\033[33m')
            for file in untracked:
                print('   ' + file)
        else:
            print('0 files found for tracking')

        print('\033[m')
    else:
        print('\033[1mDirectory without new files or changes!\033[m')
        print('\033[33m0 files found for track or commit\033[m')


def _commit(files: Union[list, str], message: str) -> None:
    author = _config.get_author()
    author_name = author.get('name')
    author_email = author.get('email')

    if not all([author_name, author_email]):
        print('\033[1;31mPlease set the author\'s name and email before\033[m')
        print('\033[33mUse "config" command to set author information\033[m')
        print('Example: \033[4mbox config --name "Name" --email "Email"\033[m')
        sys.exit(1)

    if not message:
        print('\033[1;31mA short message is required to commit\033[m')
        print('\033[33mUse "-m" flag and insert a message to commit\033[m')
        sys.exit(1)

    tracked = tracker.get_tracked()
    uncommitted = _get_uncommitted_files(tracked)

    try:
        if files == "*":
            time_s = time.time()
            commit_id = commit.commit(author_name, author_email, uncommitted, message)
            files = uncommitted
        else:
            for file in files:
                if not path.isfile(file):
                    print(f'\033[1;31mFile {repr(file)} dont\'t exists\033[m')
                    sys.exit(1)
                elif file not in tracked:
                    print(f'\033[1;31mFile {repr(file)} not tracked for commit\033[m')
                    print('\033[33mUse "add" argument to track this file\033[m')
                    sys.exit(1)

            time_s = time.time()
            commit_id = commit.commit(author_name, author_email, files, message)

        print(f'Commit #\033[4m{commit_id[:7]}\033[m "{message}"')
        print(f'\033[33m{len(files)} files committed in {time.time() - time_s:.3f}s\033[m')
    except exceptions.NoFilesToCommitError:
        print('\033[1;31mNo files changed to commit\033[m')
        print('\033[33mYou can only commit changed and tracked files\033[m')


def _log() -> None:
    commits = commit.get_commits()

    for cid, cdata in reversed(commits.items()):
        message = cdata['message']
        date = cdata['date']
        files = len(cdata['objects'])
        author = cdata['author']
        author_email = cdata['author_email']
        print(f'{files} file(s) in \033[34;4m{cid[:7]}\033[m by {author} <{author_email}> '
              f'({date}) \033[33m{repr(message)}\033[m')


def _diff() -> None:
    tracked = tracker.get_tracked()

    for file, info in tracked.items():
        if not info['binary']:
            with open(file, 'r') as reader:
                content = reader.readlines()

            merged = commit.merge_objects(file)
            current_lines = utils.enumerate_lines(content)
            diff = utils.difference_lines(merged, current_lines)
            
            if diff:
                print(f'\033[1mfile {repr(file)} diff\033[m')
                print(f'\033[33m{len(diff)} lines changed\n\033[m')

                for number, line in diff.items():
                    if line is None:
                        print(f'    \033[31m{number} | -- {merged[number]}\033[m')
                    else:
                        print(f'    \033[32m{number} | ++ {line}\033[m')

                print()


def _integrity():
    if commit.check_integrity():
        print('\033[1;32mCommits without external changes\033[m')
    else:
        print('\033[1;31mSome commits were changed inappropriately\033[m')

    print(f'\033[33m{len(commit.get_commits())} commits checked\033[m')


def main() -> None:
    parser = ArgEasy(
        name='Box',
        description='Quick and simple file versioning with Box.',
        version=__version__
    )

    parser.add_argument('init', 'Init a empty repository', action='store_true')
    parser.add_argument('status', 'View uncommitted and untracked files', action='store_true')
    parser.add_argument('log', 'View commits log', action='store_true')
    parser.add_argument('diff', 'Get difference of files', action='store_true')
    parser.add_argument('integrity', 'Check commits integrity', action='store_true')
    parser.add_argument('config', 'Global config', action='store_true')
    parser.add_argument('add', 'Add new files to track list', action='append')
    parser.add_argument('commit', 'Commit files', action='append')

    parser.add_flag('-a', 'Select all files to tracking', action='store_true')
    parser.add_flag('-am', 'Commit all changed files add insert a message')
    parser.add_flag('-m', 'A short message to commit')

    parser.add_flag('--name', 'Set author name')
    parser.add_flag('--email', 'Set author email')

    args = parser.parse()

    if not args.init and not path.isdir(OBJECTS_PATH):
        print('\033[1;31mRepository not found\033[m')
        print('\033[33mCreate a repository with "init" command\033[m')
        sys.exit(1)

    if args.init:
        _init()
    elif args.add is not None:
        if args.a:
            _add('*')
        else:
            _add(args.add)
    elif args.status:
        _status()
    elif args.log:
        _log()
    elif args.commit is not None:
        if args.am:
            _commit('*', args.am)
        else:
            _commit(args.commit, args.m)
    elif args.diff:
        _diff()
    elif args.integrity:
        _integrity()
    elif args.config:
        name, email = args.name, args.email
        if name or email:
            _config.set_author(name, email)
        else:
            print('\033[1;31mName or email is required\033[m')
            print('\033[33mUse "--name" or "--email" flag\033[m')



# Box, file versioning.
# Copyright (C) 2023  Firlast
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import json
from os import path
from hashlib import md5

from . import exceptions


class Tracker:
    def __init__(self) -> None:
        """
        Create a new instance from the Tracker class.
        """

        self._tracker_file = path.join('.box', 'tracker.json')

    @staticmethod
    def get_file_hash(filepath: str) -> str:
        with open(filepath, 'rb') as file:
            _hash = md5(file.read())

        return _hash.hexdigest()

    def dump_tracker(self, data: dict) -> None:
        with open(self._tracker_file, 'w') as tracker:
            json.dump(data, tracker, indent=2)

    def get_tracked(self) -> dict:
        """
        Get all tracked files.
        :return: Tracked files
        """

        try:
            with open(self._tracker_file) as tracker:
                tracked = json.load(tracker)
        except FileNotFoundError:
            tracked = {}

        return tracked

    def get_tracked_file(self, filepath: str) -> dict:
        tracked = self.get_tracked()
        try:
            file_tracked = tracked[filepath]
        except KeyError:
            raise exceptions.FileNotTrackedError(f'File {repr(filepath)} not tracked')

        return file_tracked

    def update_track_info(self, filepath: str, committed: bool, update_hash: bool = False) -> None:
        """
        Update `committed` status and file hash from
        tracked file.

        :param filepath: File path
        :param committed: File has committed
        :param update_hash: If true, this method update the
        file hash in tracked.
        :return:
        """

        tracked = self.get_tracked()

        if filepath in tracked:
            if update_hash:
                tracked[filepath]['hash'] = self.get_file_hash(filepath)
            tracked[filepath]['committed'] = committed
            self.dump_tracker(tracked)
        else:
            raise exceptions.FileNotTrackedError(f'File "{filepath}" not tracked')

    def track(self, files_list: list) -> None:
        """
        Track a new files.

        This method generate a hash from file and add
        this information to a tracker file specified in
        `self._tracker_file`.

        :param files_list: Files path to track
        :return: None
        """

        tracked = self.get_tracked()

        for filepath in files_list:
            try:
                with open(filepath) as _test_file:
                    _test_file.read()
            except UnicodeDecodeError:
                binary = True
            else:
                binary = False

            file_hash = self.get_file_hash(filepath)
            tracked[filepath] = dict(hash=file_hash, committed=False, binary=binary)

        self.dump_tracker(tracked)



import os
import json
import pathlib

HOME_PATH = pathlib.Path.home()
BOX_CONFIG_PATH = os.path.join(HOME_PATH, '.box.config.json')


def get_author() -> dict:
    try:
        with open(BOX_CONFIG_PATH, 'r') as file:
            config = json.load(file)
    except FileNotFoundError:
        return {}
    else:
        return config.get('author')


def set_author(name: str = None, email: str = None) -> None:
    author_info = get_author()

    if name:
        author_info['name'] = name

    if email:
        author_info['email'] = email

    with open(BOX_CONFIG_PATH, 'w') as file:
        json.dump({'author': author_info}, file, indent=2)



from setuptools import setup

from box.__init__ import __version__

with open('README.md') as file:
    readme = file.read()

setup(
    author='Firlast',
    author_email='firlastinc@gmail.com',
    name='box-vcs',
    description='Simple and fast file versioning with Box',
    version=__version__,
    packages=['box'],
    url='https://github.com/firlast/box',
    long_description=readme,
    long_description_content_type='text/markdown',
    license='GNU GPLv2',
    python_requires='>=3.7',
    install_requires=['argeasy==3.0.0'],
    entry_points={
        'console_scripts': [
            'box = box.__main__:main'
        ]
    },
    keywords=['version', 'control', 'versioning', 'vcs'],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Version Control'
    ],
)
